# vim: set fileencoding=utf-8 :
from __future__ import absolute_import, division

# vim: set fileencoding=utf-8 :
from __future__ import absolute_import, division
from dictalchemy.tests import TestCase, Named, NamedWithOtherPk,\
        NamedOtherColumnName,\
        NamedWithSynonym, OneToManyChild, OneToManyParent,\
        M2mLeft, M2mRight,\
        MultipleChildParent, MultipleChildChild1, MultipleChildChild2,\
        MultipleChildChild1Child, WithHybrid, WithDefaultInclude


class TestFromdict(TestCase):

    def test_simple(self):
        s = self.session
        named = Named('a name')
        s.add(named)
        s.commit()
        new = {'name': 'other name'}
        named.fromdict(new)
        assert named.asdict() == {'id': named.id, 'name': new['name']}

    def test_simple_aborts_on_pk(self):
        s = self.session
        named = Named('a name')
        s.add(named)
        s.commit()
        new = {'id': 7}
        try:
            named.fromdict(new)
        except:
            pass
        else:
            assert False

    def test_fromdict_allow_pk(self):
        s = self.session
        named = Named('a name')
        s.add(named)
        s.commit()
        new = {'id': 7}
        named.fromdict(new, allow_pk=True)
        assert named.asdict() == {'id': new['id'], 'name': named.name}

    def test_simple_aborts_on_named_pk_1(self):
        s = self.session
        named = NamedWithOtherPk('a name')
        s.add(named)
        s.commit()
        new = {'id': 7}
        try:
            named.fromdict(new)
            assert False
        except Exception, e:
            assert True, str(e)

    def test_one_to_many_plain(self):
        child = OneToManyChild('child')
        self.session.add(child)
        parent = OneToManyParent('parent')
        parent.child = child
        self.session.add(parent)
        self.session.commit()

        parent.fromdict({'name': 'Parent new name',\
                'child': {'name': 'Child new name'}}, follow=['child'])
        result = parent.asdict(follow=['child'])
        expected = {'id': parent.id, 'name': 'Parent new name',\
                'child': {'id': child.id, 'name': 'Child new name'}}

        assert result == expected, "%r == %r" % (result, expected)

    def test_include(self):
        h = WithHybrid(1)
        h.fromdict({'id': 17}, include=['id'])
        assert h.id == 17

    def test_default_include(self):
        h = WithDefaultInclude(1)
        h.fromdict({'id_alias': 4})
        assert h.id_alias == 4




# vim: set fileencoding=utf-8 :
""" Constants """
from __future__ import absolute_import, division

default_exclude = None
"""Default value for asdict_exclude"""
default_exclude_underscore = True
"""Default value for fromdict_allow_pk if not set"""
default_fromdict_allow_pk = False



# vim: set fileencoding=utf-8 :
"""Contains errors used by dictalchemy"""
from __future__ import absolute_import, division

class DictalchemyError(Exception):
    """Base class for Dictalchemy"""
    pass


class UnsupportedRelationError(DictalchemyError):
    """Raised when a relation is not supported by asdict or fromdict.

    The name of the relation can be accessed from 'relation_key'
    """

    def __init__(self, relation_key):
        self.relation_key = relation_key

    def __str__(self):
        return "Relation '%r' is not supported" % (self.relation_key,)


class MissingRelationError(DictalchemyError):
    """Raised when a relationship is missing

    The name of the relation can be accessed from 'relation_key'
    """

    def __init__(self, relation_key):
        self.relation_key = relation_key

    def __str__(self):
        return "Relation '%r' is not found" % (self.relation_key,)




# vim: set fileencoding=utf-8 :
from __future__ import absolute_import, division
from dictalchemy.tests import TestCase, Named, NamedOtherColumnName,\
        NamedWithSynonym, OneToManyChild, OneToManyParent,\
        M2mLeft, M2mRight,\
        MultipleChildParent, MultipleChildChild1, MultipleChildChild2,\
        MultipleChildChild1Child, WithHybrid, WithDefaultInclude,\
        WithAttributeMappedCollection, WithAttributeMappedCollectionChild


class TestAsdict(TestCase):

    def _setup_multiple_child(self):
        s = self.session
        p = MultipleChildParent('parent')
        c1 = MultipleChildChild1('child1')
        c2 = MultipleChildChild2('child2')
        p.child1 = c1
        p.child2 = c2
        s.add(p)
        s.commit()
        return p, c1, c2

    def _setup_multiple_child_child(self):
        p, c1, c2 = self._setup_multiple_child()
        s = self.session
        c1c = MultipleChildChild1Child('child1child')
        c1.child = c1c
        s.add(c1)
        s.commit()
        return p, c1, c2, c1c

    def test_dict(self):
        named = Named('a name')
        named.asdict() == dict(named)

    def test_exclude_flag(self):
        named = Named('a name')
        assert named.asdict(exclude=['id']) == {'name': 'a name'}

    def test_exclude_pk(self):
        named = Named('a name')
        self.session.add(named)
        self.session.commit()
        assert named.asdict(exclude_pk=True) == {'name': named.name}

    def test_named_without_save(self):
        named = Named('a name')
        assert named.asdict() == {'id': None, 'name': named.name}

    def test_named_with_save(self):
        named = Named('a name')
        self.session.add(named)
        self.session.commit()
        assert named.asdict() == {'id': named.id, 'name': named.name}

    def test_named_other_columnname(self):
        named = NamedOtherColumnName('a name')
        assert named.asdict() == {'id': None, 'name': named.name}

    def test_named_synonym(self):
        named = NamedWithSynonym('a name')
        assert named.asdict() == {'id': None, 'name': named.name}

    def test_one_to_many_plain(self):
        child = OneToManyChild('child')
        parent = OneToManyParent('parent')
        parent.child = child
        self.session.add(parent)
        self.session.commit()
        assert parent.asdict() == {'id': parent.id, 'name': parent.name}

    def test_one_to_many_follow(self):
        child = OneToManyChild('child')
        parent = OneToManyParent('parent')
        parent.child = child
        self.session.add(parent)
        self.session.commit()
        assert parent.asdict(follow=['child']) ==\
                {'id': parent.id, 'name': parent.name,
                        'child': child.asdict()}

    def test_many_to_many_follow(self):
        s = self.session
        l1 = M2mLeft('l1')
        r1 = M2mRight('r1')
        r2 = M2mRight('r2')
        l1.rights.append(r1)
        l1.rights.append(r2)
        s.add(l1)
        s.commit()
        assert l1.asdict(follow=['rights']) == {'id': l1.id, 'name': l1.name,
                'rights': [r1.asdict(), r2.asdict()]}
        assert r1.asdict(follow=['lefts']) == {'id': r1.id, 'name': r1.name,
                'lefts': [l1.asdict()]}

    def test_one_to_many_follow_with_arguments(self):
        s = self.session
        l1 = M2mLeft('l1')
        r1 = M2mRight('r1')
        r2 = M2mRight('r2')
        l1.rights.append(r1)
        l1.rights.append(r2)
        s.add(l1)
        s.commit()
        assert l1.asdict(follow={'rights': {'exclude': ['id']}}) ==\
                {'id': l1.id, 'name': l1.name,
                    'rights': [{'name': r1.name}, {'name': r2.name}]}
        assert r1.asdict(follow={'lefts': {'exclude': ['id']}}) ==\
                {'id': r1.id, 'name': r1.name,
                    'lefts': [{'name': l1.name}]}

    def test_multiple_child_follow_only_one(self):
        p, c1, c2 = self._setup_multiple_child()
        assert p.asdict(follow=['child1']) == {'id': p.id, 'name': p.name,
                'child1': {'id': c1.id, 'name': c1.name}}

    def test_multiple_child_follow_two_with_different_arguments(self):
        p, c1, c2 = self._setup_multiple_child()
        assert p.asdict(follow={'child1': {},
            'child2': {'exclude': ['id']}}) ==\
                    {'id': p.id, 'name': p.name,
                'child1': {'id': c1.id, 'name': c1.name},
                'child2': {'name': c2.name}}

    def test_multiple_child_follow_two_levels(self):
        p, c1, c2, c1c = self._setup_multiple_child_child()
        assert p.asdict(follow={'child1': {'follow': ['child']}}) ==\
                    {'id': p.id, 'name': p.name,
                'child1': {'id': c1.id, 'name': c1.name,
                    'child': {'id': c1c.id, 'name': c1c.name}}}

    def test_multiple_child_follow_two_levels_with_arguments(self):
        p, c1, c2, c1c = self._setup_multiple_child_child()
        assert p.asdict(follow={'child1': {'follow':\
                {'child': {'exclude': ['id']}}}}) ==\
                    {'id': p.id, 'name': p.name,
                'child1': {'id': c1.id, 'name': c1.name,
                    'child': {'name': c1c.name}}}

    def test_hybrid_property(self):
        assert WithHybrid(2).asdict(include=['id']) == {'id': 2}

    def test_default_include(self):
        assert WithDefaultInclude(2).asdict() == {'id': 2, 'id_alias': 2}

    def test_attribute_mapped_collection(self):
        p = WithAttributeMappedCollection()
        p.childs['child1'] = WithAttributeMappedCollectionChild('child1')
        assert p.asdict(follow=['childs']) == {'childs':\
                {'child1':\
                {'parent_id': None, 'id': None, 'name': 'child1'}}, 'id': None}



# vim: set fileencoding=utf-8 :
"""

~~~~~~~~~~~~
Introduction
~~~~~~~~~~~~

This library adds basic functionality for getting a dict from an SQLAlchemy
model and updating a model from a dict.

SQLAlchemy is a very complex library. It contains synonyms, lists, sets, mixins
and god knows what. Automatically convert that to a dict resp. update from a
dict is not a simple task. This library should not be used on more complex
models without thorough testing, it should however be fine to use on simple
models.

~~~~~~~~~~~~~~~~~~~~~
A note about synonyms
~~~~~~~~~~~~~~~~~~~~~

Synonyms wraps a reader and a writer method. These methods can do whatever they
want to so there is no way to safely updated data with synonyms.

~~~~~~~~~~~~~~~~~~~~~~~~~
Bugs and missing features
~~~~~~~~~~~~~~~~~~~~~~~~~

This library is developed as a helper-library to another piece of software.
What works for that software is also implemented and tested in dictalchemy.

There are bugs and missing features in this library. Please don't hesitate to
register issues at `github <https://github.com/danielholmstrom/dictalchemy>`_.

"""
from __future__ import absolute_import, division

from dictalchemy.classes import DictableModel
from dictalchemy.utils import make_class_dictable, asdict
from dictalchemy.errors import DictalchemyError, UnsupportedRelationError,\
        MissingRelationError

__all__ = [DictableModel, make_class_dictable, asdict, DictalchemyError,
        UnsupportedRelationError, MissingRelationError]



"""
~~~~~~~~~~~
Dictalchemy
~~~~~~~~~~~

Contains asdict() and fromdict() methods that will work on SQLAlchemy declarative models.

Read more in the source or on github <https://github.com/danielholmstrom/dictalchemy>.
"""

import os
from setuptools import find_packages
from distutils.core import setup

here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, 'README.md')).read()

# Requirements for the package
install_requires = ['Sphinx==1.1.3',
        'nose==1.2.1',
        'coverage==3.5.2',
        'SQLAlchemy==0.8.0b2',
        ]


# Requirement for running tests
test_requires = install_requires

setup(name='dictalchemy',
    version='0.1a4',
    description="Contains asdict and fromdict methods for SQL-Alchemy "
            "declarative models",
    long_description=__doc__,
    url='http://github.com/danielholmstrom/dictalchemy/',
    license='MIT',
    author='Daniel Holmstrom',
    author_email='holmstrom.daniel@gmail.com',
    platforms='any',
    classifiers=['Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License'],
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    install_requires=install_requires,
    tests_require=test_requires,
    test_suite='dictalchemy',
    )



# vim: set fileencoding=utf-8 :
from __future__ import absolute_import, division

from dictalchemy import utils


class DictableModel(object):
    """Adds iteration and asdict() method to an sqlalchemy class.

    :ivar dictalchemy_exclude: List of properties that should always be \
            excluded.
    :ivar dictalchemy_exclude_underscore: If True properties starting with an \
            underscore will always be excluded.
    :ivar dictalchemy_fromdict_allow_pk: If True the primary key can be \
            updated by :meth:`dictalchemy.fromdict`.

    """

    asdict = utils.asdict

    fromdict = utils.fromdict

    __iter__ = utils.iter
